package eu.unitn.disi.db.dkcore;

/**
 * Implementation of "Distributed  k-Core View Materialization and Maintenance for Large Dynamic Graphs"
 *  from Aksu, H. Dept. of Comput. Eng., Bilkent Univ., Ankara, Turkey
 *  Canim, M. ; Yuan-Chi Chang ; Korpeoglu, I. ; Ulusoy, O.
 *
 * @author Martin Brugnara (martin.brugnara@alumni.unitn.it) [<a href="http://db.disi.unitn.eu/">dbTrento</a>]
 */

import eu.unitn.disi.db.dkcore.rc.PreProcessing;
import eu.unitn.disi.db.dkcore.rc.Processing;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.hbase.HColumnDescriptor;
import org.apache.hadoop.hbase.HTableDescriptor;
import org.apache.hadoop.hbase.client.HConnection;
import org.apache.hadoop.hbase.client.HConnectionManager;
import org.apache.hadoop.hbase.client.HTableInterface;
import org.apache.hadoop.hbase.client.Put;

import java.io.*;
import java.util.ArrayList;
import java.util.regex.Pattern;

import static eu.unitn.disi.db.dkcore.Config.EMPTY_FAMILY;
import static eu.unitn.disi.db.dkcore.Config.GRAPH_TABLE_NAME;

public class DBLoader {
    private static final Log LOG = LogFactory.getLog(DBLoader.class);

    private final static String lineRegexp = "^(#.*|\\d+\\s*(\\s*\\d+\\s*)?)$";
    private final static String spacesRegexp = "\\s+";

    public boolean loadGraph(InputStream src) {
        Config config = Config.getConfig();

        HConnection connection;
        try {
            // HBase connection
            connection = HConnectionManager.createConnection(config.hbaseConf);
            HTableInterface graph = connection.getTable(GRAPH_TABLE_NAME);
            graph.setAutoFlush(false);
            // graph.getWriteBufferSize()

            // Parsing setup
            Pattern p = Pattern.compile(lineRegexp),
                    s = Pattern.compile(spacesRegexp);
            BufferedReader br = new BufferedReader(new InputStreamReader(src));
            String input;
            String currentVertex = "";

            ArrayList<Put> puts = new ArrayList<>(Config.PUT_BATCH_SIZE);

            while ((input = br.readLine()) != null) {
                if (input.isEmpty() || input.startsWith("#"))
                    continue; // it's a Comment
                else if (!p.matcher(input).matches()) {
                    LOG.warn(String.format("[EE] Malformed input: %s", input));
                    continue;
                }

                String[] row = s.split(input, 2);
                String from = row[0].trim();

                Put op;
                if (!from.equals(currentVertex)) { // Add new source vertex
                    op = Utils.EmptyPut(from);
                    currentVertex = from;
                    puts.add(op);
                }

                if (row.length == 2) { // edge
                    String to = row[1].trim();
                    op = Utils.EmptyPut(from, to);
                    puts.add(op);
                }

                if (puts.size() > Config.PUT_BATCH_SIZE) {
                    graph.put(puts);
                    puts.clear();
                }
            }

            if(!puts.isEmpty()) graph.put(puts);
            graph.flushCommits();

            graph.close();
            connection.close();

            // Early return on success
            return true;

        } catch (Exception e) {
            System.err.print(e.getMessage());
            e.printStackTrace();
        }

        // Die if we did not successfully completed the try {}
        return false;
    }

    boolean createGraphTable(InputStream regions) {
        HTableDescriptor graphTD = new HTableDescriptor(GRAPH_TABLE_NAME);
        try {
            graphTD.addFamily(new HColumnDescriptor(EMPTY_FAMILY));
            graphTD.addCoprocessor(PreProcessing.class.getName());
            graphTD.addCoprocessor(Processing.class.getName());
            return Shared.createTable(graphTD, Shared.readSplitKeys(regions));
        } catch (IOException e) {
            e.printStackTrace();
            return false;
        }
    }



    // Create graph table and load the graph.
    public static boolean load(String graphPath, String splitPointsPath) {
        try {
            InputStream graph = new FileInputStream(new File(graphPath));
            InputStream regions = new FileInputStream(new File(splitPointsPath));

            DBLoader loader = new DBLoader();

            LOG.info("Create graph table...");
            long start = System.currentTimeMillis();
            if (!loader.createGraphTable(regions)) return false;
            LOG.info(String.format("completed in %d ms\n",
                    System.currentTimeMillis() - start));

            LOG.info("Loading the graph...");
            start = System.currentTimeMillis();
            if (!loader.loadGraph(graph)) return false;
            String msg = String.format("Graph loaded in %d ms",
                    System.currentTimeMillis() - start);
            LOG.info(msg);
            System.out.println(msg);

            // early success
            return true;
        } catch (FileNotFoundException e) {
            System.err.println("Files not found.");
            e.printStackTrace();
        }

        // Fail if not completed try {}
        return false;
    }

    /*
    * 0) Graph
    * 1) n partitions
    */
    public static void main(String ... args) {
        try {
            if (!load(args[0], args[1])) LOG.error("DBLoad failed");
        } catch (IndexOutOfBoundsException e) {
            System.err.println("Not enough arguments: graph_file, region_file");
        }
    }
}
