package eu.unitn.disi.db.dkcore;

/**
 * Implementation of "Distributed  k-Core View Materialization and Maintenance for Large Dynamic Graphs"
 *  from Aksu, H. Dept. of Comput. Eng., Bilkent Univ., Ankara, Turkey
 *  Canim, M. ; Yuan-Chi Chang ; Korpeoglu, I. ; Ulusoy, O.
 *
 * @author Martin Brugnara (martin.brugnara@alumni.unitn.it) [<a href="http://db.disi.unitn.eu/">dbTrento</a>]
 */

import eu.unitn.disi.db.dkcore.rc.PreProcessing;
import eu.unitn.disi.db.dkcore.rc.PreProcessingProtocol;
import eu.unitn.disi.db.dkcore.rc.Processing;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.hbase.HColumnDescriptor;
import org.apache.hadoop.hbase.HTableDescriptor;
import org.apache.hadoop.hbase.ZooKeeperConnectionException;
import org.apache.hadoop.hbase.client.HConnection;
import org.apache.hadoop.hbase.client.HConnectionManager;
import org.apache.hadoop.hbase.client.HTableInterface;

import static org.apache.hadoop.hbase.HConstants.EMPTY_START_ROW;
import static org.apache.hadoop.hbase.HConstants.EMPTY_END_ROW;

import java.io.*;
import java.util.Map;

import static eu.unitn.disi.db.dkcore.Config.*;


public class KCore {
    private static final Log LOG = LogFactory.getLog(KCore.class);

    public static boolean compute(String splitPointsPath, int k) {
        long start = System.currentTimeMillis();

        byte[][] splitPoints = null;

        try {
            splitPoints = Shared.readSplitKeys(new FileInputStream(new File(splitPointsPath)));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
            return false;
        } catch (IOException e) {
            e.printStackTrace();
        }

        if (!createTL(splitPoints)) return false;

        Config config = Config.getConfig();
        HConnection connection;
        HTableInterface graph;

        try {
            connection = HConnectionManager.createConnection(config.hbaseConf);
            graph = connection.getTable(GRAPH_TABLE_NAME);
        } catch (ZooKeeperConnectionException e) {
            e.printStackTrace();
            return false;
        } catch (IOException e) {
            e.printStackTrace();
            return false;
        }

        try {
            LOG.info("RPC compute degree");
            graph.coprocessorExec(PreProcessingProtocol.class, EMPTY_START_ROW, EMPTY_END_ROW,
                    instance -> {
                        instance.computeDegrees();
                        return null;
                    });
        } catch (Throwable throwable) {
            throwable.printStackTrace();
            return false;
        }


        {
            try {
                LOG.info(String.format("Processing k = %d", k));
                final int finalK = k;

                LOG.info("RPC Compute QNC");
                graph.coprocessorExec(PreProcessingProtocol.class, EMPTY_START_ROW, EMPTY_END_ROW,
                        instance -> {
                            instance.computeQnc(finalK);
                            return null;
                        });

                LOG.info("Create table for export");
                if (!createKGraph(splitPoints, k)) return false;

                LOG.info("RPC Filtered Export");
                graph.coprocessorExec(PreProcessingProtocol.class, EMPTY_START_ROW, EMPTY_END_ROW,
                        instance -> {
                            instance.filteredExport(finalK);
                            return null;
                        });

                // NOTE: until here OK

                LOG.info("Base kCore");
                long iteration = 0;
                HTableInterface kGraph = connection.getTable(Utils.kGraphName(k));

                Map<byte[], Boolean> results;
                do {
                    // if ((iteration++%100) == 0) LOG.info(String.format("Iteration %d", iteration));
                    LOG.info(String.format("Iteration %d", iteration++));

                    results = kGraph.coprocessorExec(PreProcessingProtocol.class,
                            EMPTY_START_ROW, EMPTY_END_ROW,
                            instance -> instance.filterOutEdges(finalK)
                    );

                } while (results.values().contains(true));

                String msg = String.format("kCore computation completed in %d ms",
                        System.currentTimeMillis() - start);
                LOG.info(msg);
                System.out.println(msg);


                kGraph.close();
                graph.close();
                connection.close();
            } catch (Throwable throwable) {
                throwable.printStackTrace();
                return false;
            }
        }

        return true;
    }


    static boolean createTL(byte[][] splitPoints) {
        HTableDescriptor tlTD = new HTableDescriptor(TL_TABLE_NAME);

        HColumnDescriptor degreeColumn = new HColumnDescriptor(DEGREES);
        degreeColumn.setMinVersions(1);
        degreeColumn.setMaxVersions(1);
        tlTD.addFamily(degreeColumn);

        HColumnDescriptor qncColumn = new HColumnDescriptor(QNCS);
        qncColumn.setMinVersions(1);
        qncColumn.setMaxVersions(1);
        tlTD.addFamily(qncColumn);

        return Shared.createTable(tlTD, splitPoints);
    }

    static boolean createKGraph(byte[][] splitPoints, int k) {
        HTableDescriptor kGraphTD = new HTableDescriptor(Utils.kGraphName(k));
        try {
            kGraphTD.addFamily(new HColumnDescriptor(EMPTY_FAMILY));
            kGraphTD.addCoprocessor(PreProcessing.class.getName());
            kGraphTD.addCoprocessor(Processing.class.getName());
        } catch (IOException e) {
            e.printStackTrace();
            return false;
        }
        return Shared.createTable(kGraphTD, splitPoints);
    }


}
