package eu.unitn.disi.db.dkcore;

/**
 * Implementation of "Distributed  k-Core View Materialization and Maintenance for Large Dynamic Graphs"
 *  from Aksu, H. Dept. of Comput. Eng., Bilkent Univ., Ankara, Turkey
 *  Canim, M. ; Yuan-Chi Chang ; Korpeoglu, I. ; Ulusoy, O.
 *
 * @author Martin Brugnara (martin.brugnara@alumni.unitn.it) [<a href="http://db.disi.unitn.eu/">dbTrento</a>]
 */

import eu.unitn.disi.db.dkcore.rc.ProcessingProtocol;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.hbase.HBaseConfiguration;
import org.apache.hadoop.hbase.client.*;
import org.apache.hadoop.hbase.ipc.HBaseClient;
import org.apache.hadoop.hbase.util.Bytes;

import java.io.*;
import java.util.*;
import java.util.function.Function;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import static eu.unitn.disi.db.dkcore.Config.*;
import static java.util.stream.Collectors.counting;
import static java.util.stream.Collectors.groupingBy;
import static org.apache.hadoop.hbase.HConstants.EMPTY_END_ROW;
import static org.apache.hadoop.hbase.HConstants.EMPTY_START_ROW;
import static eu.unitn.disi.db.dkcore.Config.INSERT_TIMEOUT;

public class Online {
    private final static String interactionRegexp = "^(#.*|(\\+|-)\\s+\\d+\\s+\\d+\\s*)$";
    private final static String spacesRegexp = "\\s+";

    private static final Log LOG = LogFactory.getLog(KCore.class);

    protected HTableInterface graph, tl, kGraph;

    public static boolean process (String opsPath, int k) {
        try {

            InputStream ops = new FileInputStream(new File(opsPath));
            BufferedReader br = new BufferedReader(new InputStreamReader(ops));

            Pattern p = Pattern.compile(interactionRegexp),
                    s = Pattern.compile(spacesRegexp);

            String input;

            Online o = new Online();

            Config config = Config.getConfig();


            Configuration hbaseConfSmallRPC = HBaseConfiguration.create(config.hbaseConf);

            HConnection connection = HConnectionManager.createConnection(hbaseConfSmallRPC);
            hbaseConfSmallRPC.setLong("hbase.rpc.timeout", INSERT_TIMEOUT);
            hbaseConfSmallRPC.setInt("hbase.client.retries.number", 1);

            o.graph = connection.getTable(GRAPH_TABLE_NAME);
            o.tl = connection.getTable(TL_TABLE_NAME);
            o.kGraph = connection.getTable(Utils.kGraphName(k));


            long start, end;
            long mean_add = 0, mean_add_count = 0,
                    mean_del = 0, mean_del_count = 0;
            int skipped = 0;

            while ((input = br.readLine()) != null) {
                if (input.isEmpty() || input.startsWith("#"))
                    continue;
                else if (!p.matcher(input).matches()) {
                    LOG.warn(String.format("[EE] Malformed input: %s", input));
                    continue;
                }

                String[] row = s.split(input, 3);

                long u = Long.parseLong(row[1]),
                        v = Long.parseLong(row[2]);

                String nodeU = Utils.pad(u),
                        nodeV = Utils.pad(v),
                        edge = nodeU + nodeV,
                        Redge = nodeV + nodeU;

                start = end = 0;

                switch (row[0].trim()) {
                    case "+":
                        if (o.graph.exists(new Get(Bytes.toBytes(edge))))
                            LOG.warn(String.format("[EE] adding (yet exists) :%s", input));
                        else if (!o.graph.exists(new Get(Bytes.toBytes(nodeU))) ||
                                !o.graph.exists(new Get(Bytes.toBytes(nodeV))))
                            LOG.warn(String.format("[EE] adding (missing vertex[es]) :%s", input));
                        else {
                            start = System.currentTimeMillis();
                            o.graph.put(Utils.EmptyPut(edge));
                            o.graph.put(Utils.EmptyPut(Redge));

                            long degreeU = Utils.incrementDegree(o.tl, Bytes.toBytes(nodeU), 1L);
                            long degreeV = Utils.incrementDegree(o.tl, Bytes.toBytes(nodeV), 1L);

                            Utils.updateQnc(o.graph, o.tl, u, degreeU, v, k, 1L);
                            Utils.updateQnc(o.graph, o.tl, v, degreeV, u, k, 1L);

                            if (!o.insert(u, v, k)) {
                                // Too long: clean up
                                long dU = Utils.incrementDegree(o.tl, Bytes.toBytes(nodeU), -1L);
                                long dV = Utils.incrementDegree(o.tl, Bytes.toBytes(nodeV), -1L);

                                Utils.updateQnc(o.graph, o.tl, u, dU + 1, v, k, -1L);
                                Utils.updateQnc(o.graph, o.tl, v, dV + 1, u, k, -1L);

                                o.graph.delete(new Delete(Bytes.toBytes(edge)));
                                o.graph.delete(new Delete(Bytes.toBytes(Redge)));

                                // Surrender
                                continue;
                            }

                            end = System.currentTimeMillis();
                            mean_add += end - start;
                            mean_add_count++;
                        }
                        break;
                    case "-":
                        if (!o.graph.exists(new Get(Bytes.toBytes(edge))))
                            LOG.warn(String.format("[EE] removing (not exists) :%s", input));
                        else if (!o.graph.exists(new Get(Bytes.toBytes(nodeU))) ||
                                !o.graph.exists(new Get(Bytes.toBytes(nodeV))))
                            LOG.warn(String.format("[EE] removing (missing vertex[es]) :%s", input));
                        else {
                            start = System.currentTimeMillis();
                            long degreeU = Utils.incrementDegree(o.tl, Bytes.toBytes(nodeU), -1L);
                            long degreeV = Utils.incrementDegree(o.tl, Bytes.toBytes(nodeV), -1L);

                            Utils.updateQnc(o.graph, o.tl, u, degreeU + 1, v, k, -1L);
                            Utils.updateQnc(o.graph, o.tl, v, degreeV + 1, u, k, -1L);

                            o.graph.delete(new Delete(Bytes.toBytes(edge)));
                            o.graph.delete(new Delete(Bytes.toBytes(Redge)));

                            o.delete(u, v, k);

                            end = System.currentTimeMillis();
                            mean_del += end - start;
                            mean_del_count++;
                        }
                        break;
                }

                if (start == end && start == 0)
                    skipped++;
                else
                    //LOG.info(String.format("[OPS] %s %d %d in %d ms", row[0], u, v, end - start));
                    System.out.println(String.format("[OPS %d] %s %d %d in %d ms", mean_add_count + mean_del_count, row[0], u, v, end - start));
            }

            System.out.printf("+ mean %9.2f over %d ops\n", ((float) mean_add) / mean_add_count, mean_add_count);
            System.out.printf("- mean %9.2f over %d ops\n", ((float) mean_del) / mean_del_count, mean_del_count);
            LOG.info(String.format("skipped %d ops", skipped));

            // CLEAN UP
            o.graph.close();
            o.kGraph.close();
            o.tl.close();
            connection.close();

            br.close();
            ops.close();

            return true;
        } catch (Exception e) {
            e.printStackTrace();
        }

        return false;
    }

    public boolean insert(long u, long v, int k) {
        try {
            final String nodeU = Utils.pad(u), nodeV = Utils.pad(v);

            if (Utils.existsNode(u, kGraph) && Utils.existsNode(v, kGraph)) {
                kGraph.put(Utils.EmptyPut(nodeU, nodeV));
                kGraph.put(Utils.EmptyPut(nodeV, nodeU));
                return true;
            }

            long dU, dV, qncU, qncV;
            dU = dV = qncU = qncV = -1;

            {
                Get gu = new Get(Bytes.toBytes(nodeU)),
                        gv = new Get(Bytes.toBytes(nodeV));

                byte[] qncFamily = Utils.qncFamily(k);
                gu.addColumn(DEGREES, DEGREES);
                gu.addColumn(QNCS, qncFamily);
                gv.addColumn(DEGREES, DEGREES);
                gv.addColumn(QNCS, qncFamily);

                Result[] rs = tl.get(Arrays.asList(gu, gv));

                for (Result r : rs) {
                    if (r == null || r.isEmpty())
                        throw new Exception("Error retrieving meta for u-v");

                    String node = new String(r.getRow());
                    long d = Bytes.toLong(r.getValue(DEGREES, DEGREES)),
                            qnc = Bytes.toLong(r.getValue(QNCS, qncFamily));

                    if (nodeU.equals(node)) {
                        dU = d;
                        qncU = qnc;
                    } else if (nodeV.equals(node)) {
                        dV = d;
                        qncV = qnc;
                    } else {
                        throw new Exception("Something strange happened");
                    }
                }
            }

            if (dU < k || dV < k)
                return true;

            // WHY not "and" here ?
            if (qncU >= k || qncV >= k) {
                LOG.info(String.format("performInsertTraversal(%d, %d)", u, v));
                performInsertTraversal(new ArrayList<>(Collections.singleton(nodeU)), k);
            }

            return true;
        } catch (TimeOut timeout) {
            System.out.printf("+ %d %d Timeout\n", u, v);
        } catch (Throwable throwable) {
            throwable.printStackTrace();
        }
        return false;
    }

    class TimeOut extends Exception {}

    public void performInsertTraversal(List<String> insertTraversals, int k) throws Throwable {
        Set<String> C = new HashSet<>();

        long start = System.currentTimeMillis();

        while (!insertTraversals.isEmpty()) {
            if (System.currentTimeMillis() - start > INSERT_TIMEOUT) throw new TimeOut();

            final List<String> finalInsertTraversals = insertTraversals;
            Map<byte[], List<String>> qualifyingList;
            try {
                 qualifyingList = graph.coprocessorExec(
                    ProcessingProtocol.class, EMPTY_START_ROW, EMPTY_END_ROW,
                    instance -> instance.prunedTraversal(finalInsertTraversals, k));
            } catch (HBaseClient.CallTimeoutException timeout) {
                throw new TimeOut();
            }

            if (System.currentTimeMillis() - start > INSERT_TIMEOUT) throw new TimeOut();

            // Set (avoid duplicate dest node)
            List<String> edgesToAdd = qualifyingList.values().stream()
                    .flatMap(Collection::stream)
                    .filter(e -> !C.contains(e))
                    .collect(Collectors.toList());

            C.addAll(edgesToAdd);

            List<String> nextPossibleNodes = edgesToAdd.stream()
                    .map(e -> e.substring(NODE_ENCODING_LENGTH))
                    .distinct()
                    .collect(Collectors.toList());

            if (System.currentTimeMillis() - start > INSERT_TIMEOUT) throw new TimeOut();

            Map<byte[], List<String>> nextNodes;
            try {
                 nextNodes = kGraph.coprocessorExec(
                        ProcessingProtocol.class, EMPTY_START_ROW, EMPTY_END_ROW,
                        instance -> instance.fileterExistings(nextPossibleNodes));
            } catch (HBaseClient.CallTimeoutException timeout) {
                throw new TimeOut();
            }

            insertTraversals = nextNodes.values().stream().flatMap(Collection::stream).collect(Collectors.toList());
        }

        if (System.currentTimeMillis() - start > INSERT_TIMEOUT) throw new TimeOut();

        List<Put> toPut = partialKCore(C, k, start).stream().map(Utils::EmptyPut).collect(Collectors.toList());

        kGraph.put(toPut);
        kGraph.flushCommits();
    }

    public Collection<String> partialKCore(Collection<String> C, int k, long start) throws Throwable {
        int size;
        do {
            size = C.size();

            // Compute degree in C
            Map<String, Long> dynMap = C.stream()
                    .map(e -> e.substring(0, NODE_ENCODING_LENGTH))
                    .collect(groupingBy(Function.identity(), counting()));

            // Delete u,v if dC(u) < k
            List<String> toRemove = C.stream().filter(e -> {
                String node = e.substring(0, NODE_ENCODING_LENGTH);
                return dynMap.get(node) < k;
            }).collect(Collectors.toList());
            C.removeAll(toRemove);

            // Be sure to delete also v, u
            C.removeAll(toRemove.stream()
                    .map(Utils::reverse)
                    .filter(e -> !toRemove.contains(e))
                    .collect(Collectors.toList()));

            if (System.currentTimeMillis() - start > INSERT_TIMEOUT) throw new TimeOut();

        } while (size != C.size());

        // Be sure to add both u,v and v,u
        C.addAll(C.stream()
                .map(Utils::reverse)
                .filter(e -> !C.contains(e))
                .collect(Collectors.toList()));

        return C;
    }

    public void delete(long u, long v, int k) {
        try {
            final String nodeU = Utils.pad(u), nodeV = Utils.pad(v);

            if (!kGraph.exists(new Get(Bytes.toBytes(nodeU + nodeV)))) return;

            kGraph.delete(new Delete(Bytes.toBytes(nodeU + nodeV)));
            kGraph.delete(new Delete(Bytes.toBytes(nodeV + nodeU)));

            List<String> deleteList = new ArrayList<>();
            if (Utils.d(kGraph, u) < k) {
                deleteList.add(nodeU + nodeV);
                PerformDeleteTraversals(deleteList, k);
            }

            deleteList.clear();
            if (Utils.d(kGraph, v) < k) {
                deleteList.add(nodeV + nodeU);
                PerformDeleteTraversals(deleteList, k);
            }
        } catch (Throwable throwable) {
            throwable.printStackTrace();
        }
    }

    public void PerformDeleteTraversals(List<String> deleteList, int k) throws Throwable {

        while (!deleteList.isEmpty()) {
            final List<String> finalDeleteList = deleteList;
            Map<byte[], List<String>> cascadedDeletes = kGraph.coprocessorExec(
                    ProcessingProtocol.class, EMPTY_START_ROW, EMPTY_END_ROW,
                    instance -> instance.deleteEdges(finalDeleteList, k));

            deleteList = cascadedDeletes.values().stream()
                    .flatMap(Collection::stream)
                    .collect(Collectors.toList());
        }
    }
}
