package eu.unitn.disi.db.dkcore;

/**
 * Implementation of "Distributed  k-Core View Materialization and Maintenance for Large Dynamic Graphs"
 *  from Aksu, H. Dept. of Comput. Eng., Bilkent Univ., Ankara, Turkey
 *  Canim, M. ; Yuan-Chi Chang ; Korpeoglu, I. ; Ulusoy, O.
 *
 * @author Martin Brugnara (martin.brugnara@alumni.unitn.it) [<a href="http://db.disi.unitn.eu/">dbTrento</a>]
 */

import org.apache.hadoop.hbase.client.*;
import org.apache.hadoop.hbase.filter.FilterList;
import org.apache.hadoop.hbase.filter.FirstKeyOnlyFilter;
import org.apache.hadoop.hbase.filter.KeyOnlyFilter;
import org.apache.hadoop.hbase.regionserver.HRegion;
import org.apache.hadoop.hbase.util.Bytes;
import org.apache.hadoop.hbase.util.Pair;

import java.io.IOException;
import java.util.*;

import static eu.unitn.disi.db.dkcore.Config.*;
import static org.apache.hadoop.hbase.filter.FilterList.Operator.MUST_PASS_ALL;

public class Utils {

    public static String pad(long node) {
        return pad(String.valueOf(node));
    }

    public static String pad(String node) {
        return padN(node, NODE_ENCODING_LENGTH);
    }

    private final static char PADDING_CHAR = '0';
    public static String padN(String node, int len) {
        StringBuilder b = new StringBuilder(len);
        for (int i=0; i < len - node.length(); i++) b.append(PADDING_CHAR);
        b.append(node);
        return b.toString();
    }

    public static Put EmptyPut(String node) {
        return EmptyPut(Bytes.toBytes(pad(node)));
    }

    public static Put EmptyPut(String src, String dst) {
        return EmptyPut(Bytes.toBytes(pad(src) + pad(dst)));
    }

    public static Put EmptyPut(byte key[]) {
        Put p = new Put(key);
        p.add(EMPTY_FAMILY, null, EMPTY);
        return p;
    }

    public static byte[] qncFamily(int k) {
        return Bytes.toBytes(k);
    }

    public static String kGraphName(int k) {
        return String.format("%s_%d", GRAPH_TABLE_NAME, k);
    }

    public static boolean existsNode(long node, HTableInterface table) throws IOException {
        return existsNode(Utils.pad(node), table);
    }

    public static boolean existsNode(String node, HTableInterface table) throws IOException {
        Scan scan = new Scan();
        scan.setFilter(new FilterList(MUST_PASS_ALL, new FirstKeyOnlyFilter(), new KeyOnlyFilter()));
        scan.setCaching(1);
        scan.setBatch(1);

        byte [] bnode = Bytes.toBytes(node);
        scan.setStartRow(bnode);


        try (ResultScanner scanner = table.getScanner(scan)) {
            Result result = scanner.next();
            return result != null && !result.isEmpty() && new String(result.getRow()).startsWith(node);
        }
    }


    public static HashMap<String, List<String>> groupByRegion(Pair<byte[][],byte[][]> keys, List<String> edges) {
        HashMap<String, List<String>> groups = new HashMap<>();
        ArrayList<String> startKeys = new ArrayList<>();

        for (byte[] b : keys.getFirst()) {
            String key = new String(b);
            groups.put(key, new ArrayList<>());
            startKeys.add(key);
        }

        Collections.sort(startKeys);

        System.out.println(startKeys);

        for (String edge : edges) {
            int r = Collections.binarySearch(startKeys, edge);
            r = r >= 0 ? r : (Math.abs(r)-2);
            groups.get(startKeys.get(r)).add(edge);
        }

        return groups;
    }



    public static long d(HTableInterface table, String node) throws IOException {
        return d(table, Long.valueOf(node, 10));
    }


    public static long d(HTableInterface table, long node) throws IOException {
        String srcNode = pad(node);
        Scan scan = new Scan();
        scan.setFilter(new FilterList(MUST_PASS_ALL, new FirstKeyOnlyFilter(), new KeyOnlyFilter()));
        // scan.setBatch(SCANNER_BATCH);
        scan.setCaching(PLAUSIBLE_DEGREE);
        scan.setStartRow(Bytes.toBytes(srcNode + pad(0)));
        scan.setStopRow(Bytes.toBytes(pad(node + 1)));

        long degree=0;

        Result r;
        try (ResultScanner scanner = table.getScanner(scan)) {
            while ((r = scanner.next()) != null && !r.isEmpty())
                degree++;
        }

        return degree;
    }

    public static long incrementDegree(HTableInterface tl, byte [] row, long v) throws Exception {
        Increment i = new Increment(row);
        i.addColumn(DEGREES, DEGREES, v);
        Result r = tl.increment(i);
        if (r == null || r.isEmpty()) throw new Exception("Degree: new value was not returned");
        return Bytes.toLong(r.getValue(DEGREES, DEGREES));
    }

    public static void updateQnc(HTableInterface graph, HTableInterface tl,
                                 long u, long degreeU, long v, int k, long delta) throws Exception
    {
        if (degreeU < k) return; // Nothing to do here

        byte [] qncFamily = qncFamily(k);

        if (degreeU > k) {
            // I was yet valid for QNC count
            // update only other end of edge count
            Increment g = new Increment(Bytes.toBytes(pad(v)));
            g.addColumn(QNCS, qncFamily, delta);
            tl.increment(g);
            return;
        }

        // degree == k
        // Now I am valid for QNC count
        // update all neighbors

        String srcNode = pad(u);
        Scan scan = new Scan();
        scan.setFilter(new FilterList(MUST_PASS_ALL, new FirstKeyOnlyFilter(), new KeyOnlyFilter()));
        // scan.setBatch(SCANNER_BATCH);
        scan.setCaching(PLAUSIBLE_DEGREE);
        scan.setStartRow(Bytes.toBytes(srcNode + pad(0))); // DO NOT include std alone node
        scan.setStopRow(Bytes.toBytes(pad(u + 1)));


        List<Row> incs = new ArrayList<>();
        try (ResultScanner scanner = graph.getScanner(scan)) {
            for (Result r : scanner) {
                String w = new String(r.getRow()).substring(NODE_ENCODING_LENGTH);
                Increment g = new Increment(Bytes.toBytes(w));
                g.addColumn(QNCS, qncFamily, delta);
                incs.add(g);
            }
        }

        tl.batch(incs);
    }

    public static boolean inRegion(HRegion region, String x) {
        String start = new String(region.getStartKey()),
                end = new String(region.getEndKey());

        return (start.isEmpty() || start.compareTo(x) <= 0) &&
                (end.isEmpty() || x.compareTo(end) < 0);
    }

    public static String reverse(String edge) {
        return edge.substring(NODE_ENCODING_LENGTH) +
                edge.substring(0, NODE_ENCODING_LENGTH);
    }
}
