package eu.unitn.disi.db.dkcore.rc;

/**
 * Implementation of "Distributed  k-Core View Materialization and Maintenance for Large Dynamic Graphs"
 *  from Aksu, H. Dept. of Comput. Eng., Bilkent Univ., Ankara, Turkey
 *  Canim, M. ; Yuan-Chi Chang ; Korpeoglu, I. ; Ulusoy, O.
 *
 * @author Martin Brugnara (martin.brugnara@alumni.unitn.it) [<a href="http://db.disi.unitn.eu/">dbTrento</a>]
 */

import eu.unitn.disi.db.dkcore.Config;
import eu.unitn.disi.db.dkcore.Utils;
import org.apache.hadoop.hbase.KeyValue;
import org.apache.hadoop.hbase.client.*;
import org.apache.hadoop.hbase.coprocessor.BaseEndpointCoprocessor;
import org.apache.hadoop.hbase.coprocessor.RegionCoprocessorEnvironment;
import org.apache.hadoop.hbase.filter.FilterList;
import org.apache.hadoop.hbase.filter.FirstKeyOnlyFilter;
import org.apache.hadoop.hbase.filter.KeyOnlyFilter;
import org.apache.hadoop.hbase.regionserver.HRegion;
import org.apache.hadoop.hbase.regionserver.RegionScanner;
import org.apache.hadoop.hbase.util.Bytes;
import org.apache.hadoop.hbase.util.Pair;

import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static eu.unitn.disi.db.dkcore.Config.*;
import static eu.unitn.disi.db.dkcore.Utils.*;
import static org.apache.hadoop.hbase.filter.FilterList.Operator.MUST_PASS_ALL;

public class PreProcessing extends BaseEndpointCoprocessor implements PreProcessingProtocol {

    @Override
    public void computeDegrees() throws IOException {
        Scan scan = new Scan();
        scan.setFilter(new FilterList(MUST_PASS_ALL, new FirstKeyOnlyFilter(), new KeyOnlyFilter()));
        scan.setCaching(SCANNER_CACHE);
        scan.setBatch(SCANNER_BATCH);

        HRegion region = ((RegionCoprocessorEnvironment) getEnvironment()).getRegion();

        Config config = Config.getConfig();
        HConnection connection = HConnectionManager.createConnection(config.hbaseConf);
        HTableInterface tl = connection.getTable(TL_TABLE_NAME);
        tl.setAutoFlush(false);

        RegionScanner scanner = null;
        try {
            scanner = region.getScanner(scan);

            String currentNode = null;
            long degree = 0;
            List<Put> puts = new ArrayList<>(PUT_BATCH_SIZE);

            boolean hasMore = true;
            List<KeyValue> results = new ArrayList<>(SCANNER_BATCH);
            while (hasMore) {
                hasMore = scanner.next(results);
                // for each ror
                for (KeyValue k : results) {
                    String key = new String(k.getRow());

                    if (key.length() == NODE_ENCODING_LENGTH) {
                        if (currentNode != null) puts.add(saveDegree(currentNode, degree));
                        currentNode = key;
                        degree = 0l;
                    } else {
                        degree++;
                    }
                }

                results.clear();
                if (puts.size() > PUT_BATCH_SIZE) {
                    // NOTE: we are not really respecting PUT_BATCH_SIZE but only the constraint >
                    tl.put(puts);
                    puts.clear();
                }
            }

            puts.add(saveDegree(currentNode, degree));
            tl.put(puts);
            tl.flushCommits();

            tl.close();
            connection.close();
        } finally {
            if (scanner != null)
                scanner.close();
        }
    }

    private Put saveDegree(String node, long degree) {
        Put p = new Put(Bytes.toBytes(node));
        p.add(DEGREES, DEGREES, Bytes.toBytes(degree));
        return p;
    }

    @Override
    public void computeQnc(int k) throws IOException {
        Scan scan = new Scan();
        scan.setFilter(new FilterList(MUST_PASS_ALL, new FirstKeyOnlyFilter(), new KeyOnlyFilter()));
        scan.setCaching(SCANNER_CACHE);
        scan.setBatch(SCANNER_BATCH);

        HRegion region = ((RegionCoprocessorEnvironment) getEnvironment()).getRegion();

        Config config = Config.getConfig();
        HConnection connection = HConnectionManager.createConnection(config.hbaseConf);
        HTableInterface tl = connection.getTable(TL_TABLE_NAME);
        tl.setAutoFlush(false);

        RegionScanner scanner = null;
        try {
            scanner = region.getScanner(scan);

            HashMap<String, ArrayList<String>> adjMap = new HashMap<>(); // In ram adj
            HashSet<String> bulkDegree = new HashSet<>();
            List<Put> qncBulkPut = new ArrayList<>();

            String currentNode = null;

            boolean hasMore = true;
            List<KeyValue> results = new ArrayList<>(SCANNER_BATCH);
            while (hasMore) {
                hasMore = scanner.next(results);
                // for each ror
                for (KeyValue kv : results) {
                    String key = new String(kv.getRow());

                    if (key.length() == NODE_ENCODING_LENGTH) {
                        if (currentNode != null) {
                            // We have completed iteration over a node
                            if (bulkDegree.size() > DEGREE_BULK_GET_LIMIT) { // DO we really need this here ? or use just later ?
                                localQnc(tl, k, adjMap, bulkDegree, qncBulkPut);
                                adjMap.clear();
                                bulkDegree.clear();
                            }
                        }

                        // new node
                        currentNode = key;
                        adjMap.put(currentNode, new ArrayList<>());

                    } else {
                        // outgoing edge
                        String v = key.substring(NODE_ENCODING_LENGTH);
                        adjMap.get(currentNode).add(v);
                        bulkDegree.add(v);
                    }
                }

                // Manage local iteration
                results.clear();

                if (qncBulkPut.size() > PUT_BATCH_SIZE) {
                    tl.put(qncBulkPut);
                    qncBulkPut.clear();
                }
            }

            localQnc(tl, k, adjMap, bulkDegree, qncBulkPut);
            tl.put(qncBulkPut);
            tl.flushCommits();

            tl.close();
            connection.close();
        } finally {
            if (scanner != null)
                scanner.close();
        }
    }

    private void localQnc(HTableInterface tl, int k, HashMap<String,
                     ArrayList<String>> adjMap, HashSet<String> bulkDegree,
                     Collection<Put> qncBulkPut) throws IOException {
        byte[] kColumn = qncFamily(k);

        Map<String, Long> degreeMap = Stream.of(tl.get(
                bulkDegree.stream().map(node -> {
                    Get g = new Get(Bytes.toBytes(node));
                    g.addColumn(DEGREES, DEGREES);
                    return g;
                }).collect(Collectors.toList())))
        .map(r -> {
            KeyValue kv = r.getColumnLatest(DEGREES, DEGREES);
            return new Pair<>(new String(kv.getRow()), Bytes.toLong(kv.getValue()));
        }).collect(Collectors.toMap(Pair::getFirst, Pair::getSecond));

        for (Map.Entry<String, ArrayList<String>> entry : adjMap.entrySet()) {
            Put p = new Put(Bytes.toBytes(entry.getKey()));
            p.add(QNCS, kColumn,
                    Bytes.toBytes(entry.getValue().stream().map(degreeMap::get).filter(d -> d >= k).count()));
            qncBulkPut.add(p);
        }
    }

    @Override
    public void filteredExport(int k) throws IOException {
        Scan scan = new Scan();
        scan.setFilter(new FilterList(MUST_PASS_ALL, new FirstKeyOnlyFilter(), new KeyOnlyFilter()));
        scan.setCaching(SCANNER_CACHE);
        scan.setBatch(SCANNER_BATCH);

        HRegion region = ((RegionCoprocessorEnvironment) getEnvironment()).getRegion();

        Config config = Config.getConfig();
        HConnection connection = HConnectionManager.createConnection(config.hbaseConf);
        HTableInterface tl = connection.getTable(TL_TABLE_NAME);
        HTableInterface kGraph = connection.getTable(kGraphName(k));
        kGraph.setAutoFlush(false);

        RegionScanner scanner = null;
        try {
            scanner = region.getScanner(scan);

            HashMap<String, ArrayList<String>> adjMap= new HashMap<>(); // In ram adj
            Set<String> qncBulkGet = new HashSet<>();
            List<Put> exportBulkPut = new ArrayList<>();

            String currentNode = null;

            boolean hasMore = true;
            List<KeyValue> results = new ArrayList<>(SCANNER_BATCH);
            while (hasMore) {
                hasMore = scanner.next(results);
                // for each ror
                for (KeyValue kv : results) {
                    String key = new String(kv.getRow());
                    if (key.length() == NODE_ENCODING_LENGTH) {
                        // new node: consider processing the backlog..
                        if (currentNode != null)
                            if (qncBulkGet.size() > DEGREE_BULK_GET_LIMIT) {
                                localExport(tl, k, adjMap, qncBulkGet, exportBulkPut);
                                adjMap.clear();
                                qncBulkGet.clear();
                            }

                        // new node
                        currentNode = key;
                        adjMap.put(currentNode, new ArrayList<>());
                        qncBulkGet.add(currentNode);
                    } else {
                        // outgoing edge
                        String v = key.substring(NODE_ENCODING_LENGTH);
                        adjMap.get(currentNode).add(v);
                        qncBulkGet.add(v);
                    }
                }

                // Manage local iteration
                results.clear();

                if (exportBulkPut.size() > PUT_BATCH_SIZE) {
                    // Save
                    kGraph.put(exportBulkPut);
                    exportBulkPut.clear();
                }
            }

            localExport(tl, k, adjMap, qncBulkGet, exportBulkPut);
            if (!exportBulkPut.isEmpty()) kGraph.put(exportBulkPut);
            kGraph.flushCommits();

            tl.close();
            kGraph.close();
            connection.close();
        } finally {
            if (scanner != null)
                scanner.close();
        }
    }

    private void localExport(HTableInterface tl, int k, HashMap<String,
                             ArrayList<String>> adjMap, Set<String> bulkQnc,
                             Collection<Put> exportBulkPut) throws IOException {
        byte[] kColumn = qncFamily(k);

        Map<String, Long> qncMap = Stream.of(tl.get(bulkQnc.stream().map(node -> {
            Get g = new Get(Bytes.toBytes(node));
            g.addColumn(QNCS, kColumn);
            return g;
        }).collect(Collectors.toList()))).map(r -> {
            KeyValue kv = r.getColumnLatest(QNCS, kColumn);
            return new Pair<>(new String(kv.getRow()), Bytes.toLong(kv.getValue()));
        }).collect(Collectors.toMap(Pair::getFirst, Pair::getSecond));

        for (Map.Entry<String, ArrayList<String>> e : adjMap.entrySet()) {
            if (qncMap.get(e.getKey()) < k) continue;
            String u = e.getKey();
            for (String v : e.getValue()) {
                if (qncMap.get(v) < k) continue;
                exportBulkPut.add(Utils.EmptyPut(u + v));
            }
        }
    }

    @Override
    public boolean filterOutEdges(int k) throws IOException {
        Scan scan = new Scan();
        scan.setFilter(new FilterList(MUST_PASS_ALL, new FirstKeyOnlyFilter(), new KeyOnlyFilter()));
        scan.setCaching(SCANNER_CACHE);
        scan.setBatch(SCANNER_BATCH);

        Config config = Config.getConfig();
        HConnection connection = HConnectionManager.createConnection(config.hbaseConf);
        HTableInterface kGraph = connection.getTable(kGraphName(k));

        kGraph.setAutoFlush(false);

        HRegion region = ((RegionCoprocessorEnvironment) getEnvironment()).getRegion();

        RegionScanner scanner = null;

        boolean anyDeletion = false;

        try {
            scanner = region.getScanner(scan);

            List<String> adj = new ArrayList<>();
            List<Delete> deletions = new ArrayList<>();
            String currentNode = null;

            boolean hasMore = true;
            List<KeyValue> results = new ArrayList<>(SCANNER_BATCH);

            while (hasMore) {
                hasMore = scanner.next(results);
                for (KeyValue kv : results) {
                    String key = new String(kv.getRow());
                    String u = key.substring(0, NODE_ENCODING_LENGTH), v = key.substring(NODE_ENCODING_LENGTH);
                    if (currentNode == null) currentNode = u;
                    if (!u.equals(currentNode)) {
                        anyDeletion |= deleteIfReq(region, currentNode, adj, k, deletions);

                        if (deletions.size() > PUT_BATCH_SIZE) {
                            kGraph.delete(deletions);
                            deletions.clear();
                        }

                        adj.clear();
                        currentNode = u;
                    }
                    adj.add(v);
                }
                results.clear();
            }
            anyDeletion |= deleteIfReq(region, currentNode, adj, k, deletions);
            kGraph.delete(deletions);

            kGraph.flushCommits();
            kGraph.close();
            connection.close();

        } finally {
            if (scanner != null)
                scanner.close();
        }

        return anyDeletion;
    }

    private boolean deleteIfReq(HRegion region, final String currentNode,
                                List<String> adj, int k, List<Delete> d) throws IOException {
        if (0 == adj.size() || adj.size() >= k) return false;

        for (String n: adj) {
            region.delete(new Delete(Bytes.toBytes(currentNode + n)), false);
            d.add(new Delete(Bytes.toBytes(n + currentNode)));
        }

        return true;
    }
}
