package eu.unitn.disi.db.dkcore.rc;

/**
 * Implementation of "Distributed  k-Core View Materialization and Maintenance for Large Dynamic Graphs"
 *  from Aksu, H. Dept. of Comput. Eng., Bilkent Univ., Ankara, Turkey
 *  Canim, M. ; Yuan-Chi Chang ; Korpeoglu, I. ; Ulusoy, O.
 *
 * @author Martin Brugnara (martin.brugnara@alumni.unitn.it) [<a href="http://db.disi.unitn.eu/">dbTrento</a>]
 */

import eu.unitn.disi.db.dkcore.Config;
import eu.unitn.disi.db.dkcore.Utils;
import org.apache.hadoop.hbase.KeyValue;
import org.apache.hadoop.hbase.client.*;
import org.apache.hadoop.hbase.coprocessor.BaseEndpointCoprocessor;
import org.apache.hadoop.hbase.coprocessor.RegionCoprocessorEnvironment;
import org.apache.hadoop.hbase.filter.FilterList;
import org.apache.hadoop.hbase.filter.FirstKeyOnlyFilter;
import org.apache.hadoop.hbase.filter.KeyOnlyFilter;
import org.apache.hadoop.hbase.regionserver.HRegion;
import org.apache.hadoop.hbase.regionserver.RegionScanner;
import org.apache.hadoop.hbase.util.Bytes;
import org.apache.hadoop.hbase.util.Pair;

import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static eu.unitn.disi.db.dkcore.Config.*;
import static org.apache.hadoop.hbase.filter.FilterList.Operator.MUST_PASS_ALL;

public class Processing extends BaseEndpointCoprocessor implements ProcessingProtocol {

    // TB called on graph
    @Override
    public List<String> prunedTraversal(List<String> insertTraversal, int k) throws IOException {
        HRegion region = ((RegionCoprocessorEnvironment) getEnvironment()).getRegion();
        insertTraversal.removeIf(s -> !Utils.inRegion(region, s));
        if (insertTraversal.isEmpty()) return Collections.EMPTY_LIST;

        List<String> returnList = new ArrayList<>();

        HashMap<String, ArrayList<String>> adjMap = new HashMap<>();
        HashSet<String> bulkGet = new HashSet<>();

        Scan scan = new Scan();
        scan.setFilter(new FilterList(MUST_PASS_ALL, new FirstKeyOnlyFilter(), new KeyOnlyFilter()));
        scan.setCaching(PLAUSIBLE_DEGREE);
        scan.setBatch(SCANNER_BATCH);

        // Do not mess up with RegionScanner
        insertTraversal.sort(String::compareTo);

        Config config = Config.getConfig();
        HConnection connection = HConnectionManager.createConnection(config.hbaseConf);
        HTableInterface tl = connection.getTable(TL_TABLE_NAME);

        String zero = Utils.pad(0);
        try (RegionScanner scanner = region.getScanner(scan)) {
            List<KeyValue> results = new ArrayList<>(SCANNER_BATCH);

            for (String u : insertTraversal) {
                scanner.reseek(Bytes.toBytes(u + zero));

                boolean hasMore;
                ArrayList<String> neighbors = new ArrayList<>(PLAUSIBLE_DEGREE);
                do {
                    hasMore = scanner.next(results);
                    for (KeyValue kv : results) {
                        String key = new String(kv.getRow());

                        if (key.length() == NODE_ENCODING_LENGTH) {
                            hasMore = false;
                            break;
                        }

                        neighbors.add(key.substring(NODE_ENCODING_LENGTH));
                    }
                    results.clear();
                } while (hasMore);

                adjMap.put(u, neighbors);
                bulkGet.addAll(neighbors);


                if (bulkGet.size() > DEGREE_BULK_GET_LIMIT) {
                    localPruneTraversal(tl, k, adjMap, bulkGet, returnList);
                    adjMap.clear();
                    bulkGet.clear();
                }
            }

            if (!bulkGet.isEmpty()) localPruneTraversal(tl, k, adjMap, bulkGet, returnList);
        } catch (Exception e) {
            e.printStackTrace();
        }

        tl.close();
        connection.close();

        return returnList;
    }

    private void localPruneTraversal(HTableInterface tl, int k,
                                     HashMap<String, ArrayList<String>> adjMap,
                                     HashSet<String> bulkDegree, Collection<String> returnList) throws IOException
    { // TODO: can be optimized in space
        // We could only check for degree since qnc is always <= degree.
        final byte [] qncFamily = Utils.qncFamily(k);

        Map<String, Pair<Long, Long>> tlMap = Utils.groupByRegion(
                ((HTable) tl).getStartEndKeys(), new ArrayList<>(bulkDegree))
            .entrySet().stream()
            .map(x -> {
                try {
                    return tl.get(x.getValue().stream()
                            .map(node -> {
                                Get g = new Get(Bytes.toBytes(node));
                                g.addColumn(DEGREES, DEGREES);
                                g.addColumn(QNCS, qncFamily);
                                return g;
                            }).collect(Collectors.toList()));
                } catch (IOException e) {
                    e.printStackTrace();
                    return null;
                }
            })
            .flatMap(Stream::of)
            .map(r -> {
                long degree = Bytes.toLong(r.getValue(DEGREES, DEGREES)),
                        qnc = Bytes.toLong(r.getValue(QNCS, qncFamily));
                return new Pair<>(new String(r.getRow()), new Pair<>(degree, qnc));
            }).collect(Collectors.toMap(Pair::getFirst, Pair::getSecond));

        for (Map.Entry<String, ArrayList<String>> entry : adjMap.entrySet()) {
            String src = entry.getKey();
            for (String w: entry.getValue()) {
                Pair<Long, Long> tlEntry = tlMap.get(w);
                if (tlEntry.getFirst() >= k && tlEntry.getSecond() >= k)
                    returnList.add(src + w);
            }
        }
    }


    // TB called on kGraph
    @Override
    public List<String> deleteEdges(List<String> deleteList, int k) throws IOException {
        HRegion region = ((RegionCoprocessorEnvironment) getEnvironment()).getRegion();
        deleteList.removeIf(s -> !Utils.inRegion(region, s));
        if (deleteList.isEmpty()) return Collections.EMPTY_LIST;

        Scan scan = new Scan();
        scan.setFilter(new FilterList(MUST_PASS_ALL, new FirstKeyOnlyFilter(), new KeyOnlyFilter()));
        scan.setCaching(PLAUSIBLE_DEGREE);
        scan.setBatch(SCANNER_BATCH);

        List<String> cascadeDeletes = new ArrayList<>();



        // Allocate only once pls.
        List<KeyValue> results = new ArrayList<>(SCANNER_BATCH);

        for (String edge : deleteList) {
            region.delete(new Delete(Bytes.toBytes(edge)), false);
            localDeleteEdges(region, scan, edge.substring(0, NODE_ENCODING_LENGTH), k, results, cascadeDeletes);
        }

        return cascadeDeletes;
    }

    void localDeleteEdges(HRegion region, Scan scan,
                          String node, int k, List<KeyValue> results,
                          Collection<String> cascadeDeletes) throws IOException {
        scan.setStartRow(Bytes.toBytes(node));
        try (RegionScanner scanner = region.getScanner(scan)) {
            ArrayList<String> adj = new ArrayList<>();
            boolean hasMore = true;
            while (hasMore) {
                hasMore = scanner.next(results);
                for (KeyValue kv : results) {
                    String edge = new String(kv.getRow());
                    if (!edge.startsWith(node)) {
                        hasMore = false;
                        break;
                    }
                    adj.add(edge.substring(NODE_ENCODING_LENGTH));
                }
                results.clear();
            }

            if (adj.size() < k) {
                for (String w : adj) {
                    region.delete(new Delete(Bytes.toBytes(node + w)), false);
                    cascadeDeletes.add(w + node);
                }
            }
        }
    }

    @Override
    public List<String> fileterExistings(List<String> nodes) throws IOException {
        HRegion region = ((RegionCoprocessorEnvironment) getEnvironment()).getRegion();
        nodes.removeIf(n -> !Utils.inRegion(region, n));
        nodes.sort(String::compareTo);
        if (nodes.isEmpty()) return Collections.EMPTY_LIST;

        Scan scan = new Scan();
        scan.setFilter(new FilterList(MUST_PASS_ALL, new FirstKeyOnlyFilter(), new KeyOnlyFilter()));
        scan.setCaching(1);
        scan.setBatch(1);
        scan.setMaxVersions(1);

        List<String> filtered = new ArrayList<>();

        try (RegionScanner scanner = region.getScanner(scan)) {
            KeyValue result = null;
            List<KeyValue> results = new ArrayList<>(1);
            for (String node : nodes) {
                scanner.reseek(Bytes.toBytes(node));
                scanner.next(results);
                if (results.size() == 0 ||  // end of region
                    (result = results.get(0)) == null || // no values
                    result.isEmptyColumn() || // no values
                    !(new String(result.getRow()).startsWith(node))) // next node (not the one we are looking 4)
                    filtered.add(node);
            }
        }

        return filtered;
    }
}
